!! init.f90 written by Hy Trac
!! Last modified on Apr 14, 2005
!! dist_init.f90 Slab Decomposition Parallelized: Hugh Merz Jul 05, 2005
!! Compile with: mpif77 -fpp -g -O3 dist_init.f90 -o dist_init  -L/home/merz/lib/fftw/lib -I/home/merz/lib/fftw/include -lrfftw_mpi -lrfftw -lfftw_mpi -lfftw -lm -ldl
!! Power spectra have been checked

program dist_init 
  implicit none
  include 'mpif.h'

  character(*), parameter :: ic_path        = '/mnt/s0/merz/pmfast/100Mpc/ic/' 
  integer(4),   parameter :: nodes          = 28              !number of cpus total 
  integer(4),   parameter :: nc             = 3248            !number of total fine mesh cells
  integer(4),   parameter :: cpu_node       = 4

  logical, parameter :: write_dm = .true.
  logical, parameter :: generate_seeds = .true.
  logical, parameter :: normalize_noise = .false.
  logical, parameter :: correct_kernel= .true.

  !! Cosmo parameters
  real, parameter :: box= 100.0 !Mpc/h
  real, parameter :: redshift= 100.0
  real, parameter :: scalefactor=1/(1+redshift)
  real, parameter :: ns=1.0
  real, parameter :: s8=0.9
  real, parameter :: omegal=0.73
  real, parameter :: omegam=1.0-omegal !0.3

  real, parameter :: pi=3.14159
  real(4), parameter :: eps = 1.0e-03

  !! nk is the length of the initial power spectrum file
  integer, parameter      :: nk=361 !406
  character(*), parameter :: fntf='cmbfast.lcdm'

  !! IOform should be 'binary' or 'unformatted'
  character(*), parameter :: IOform='binary'

  !! nc is the number of cells per box length
  integer, parameter :: hc=nc/2
  real, parameter    :: ncr=nc
  real, parameter    :: hcr=hc

  !! np is the number of particles
  integer, parameter :: np=hc
  !! np should be set to nc (1:1), hc (1:2), or qc (1:4)
  real, parameter    :: npr=np
  !! mp is the mass of each particle 
  real, parameter       :: mp=(ncr/np)**3

  !! internal parallelization parameters
  integer(4), parameter :: nc_slab = nc/nodes 
  integer(4), parameter :: np_slab = np/nodes
  integer(4), parameter :: np_buffer = 4*np**2
  integer(4), parameter :: max_np = np*np*np_slab + np_buffer

! :: simulation variables
 
  integer(4) :: rank,ierr,np_local,scc,scr,scci
  integer(8) :: plan,iplan
  logical :: firstfftw

  !! Dark matter arrays
  real, dimension(6,max_np) :: xvp
  real, dimension(6,np,np,np_slab) :: xvp_grid
  real, dimension(6,np_buffer) :: xp_buf
  real, dimension(6*np_buffer) :: send_buf, recv_buf

  !! Power spectrum arrays
  real, dimension(5,nk) :: tf
  real, dimension(2,nc) :: pkm,pkdm

  !! Fourier transform arrays
  real, dimension(nc+2,nc,nc_slab) :: slab, slab_work

  !! arrays
  real, dimension(nc+2,nc,nc_slab)     :: delta
  real, dimension(nc,nc,-1:nc_slab+2) :: phi

  !! Equivalence arrays to save memory
  equivalence (xvp,xvp_grid,delta)
  equivalence (phi,slab_work)

  !! Common block
  common /aba/ phi
  common /abb/ slab
  common /abc/ xp_buf
  common /abd/ recv_buf
  common /abe/ send_buf
  common /abf/ tf
  common /abg/ pkm
  common /abh/ pkdm
  common /abi/ delta

!-------------------------------------------------------------------!
! Main Program

  call mpi_initialize
  if (rank == 0) call writeparams
  call system_clock(count=scci)
  call initvar
  call system_clock(count=scc,count_rate=scr)
  if (rank==0) print *,'done initvar:',real(scc-scci)/real(scr)
  call transferfnc
  call system_clock(count=scc,count_rate=scr)
  if (rank==0) print *,'done transferfnc:',real(scc-scci)/real(scr)
  call noisemap
  call system_clock(count=scc,count_rate=scr)
  if (rank==0) print *,'done noisemap:',real(scc-scci)/real(scr)
  call deltafield
  call system_clock(count=scc,count_rate=scr)
  if (rank==0) print *,'done deltafield:',real(scc-scci)/real(scr)
  call potentialfield
  call system_clock(count=scc,count_rate=scr)
  if (rank==0) print *,'done potentialfield:',real(scc-scci)/real(scr)
  call dm
  call system_clock(count=scc,count_rate=scr)
  if (rank==0) print *,'done dm:',real(scc-scci)/real(scr)
  if (write_dm) call writedm
  if (rank == 0) call writepowerspectra
  call system_clock(count=scc,count_rate=scr)
  if (rank==0) print *,'finished main:',real(scc-scci)/real(scr)
  call fftw(0)
  call mpi_finalize(ierr)

contains

!!------------------------------------------------------------------!!

  subroutine initvar
    implicit none
    integer :: k

    !! Initialize large arrays 
    phi=0.0
    slab=0.0
    xp_buf=0.0
    send_buf=0.0
    recv_buf=0.0
    tf=0.0
    pkm=0.0
    pkdm=0.0
    delta=0.0
 
    !! Initialize fftw so that it generates the plans
    firstfftw=.true.

    return
  end subroutine initvar

!-------------------------------------------------------------------!

  subroutine mpi_initialize
    implicit none
    
    integer(4) :: i, j, nodes_returned
  
!! set up global mpi communicator

    call mpi_init(ierr)
    if (ierr /= mpi_success) call mpi_abort(mpi_comm_world,ierr,ierr)

    call mpi_comm_size(mpi_comm_world,nodes_returned,ierr)
    if (ierr /= mpi_success) call mpi_abort(mpi_comm_world,ierr,ierr)
    if (nodes_returned /= nodes ) then
      write(*,*) 'dist_init compiled for a different number of nodes'
      write(*,*) 'mpirun nodes=',nodes_returned,'dist_init nodes=',nodes 
      call mpi_abort(mpi_comm_world,ierr,ierr)
    endif
    if (mod(nc,nodes) /= 0) then
      write(*,*) 'cannot evenly decompose mesh into slabs'
      write(*,*) 'nc=',nc,'nodes=',nodes,'mod(nc,nodes) != 0'
      call mpi_abort(mpi_comm_world,ierr,ierr)
    endif

    call mpi_comm_rank(mpi_comm_world,rank,ierr)
    if (ierr /= mpi_success) call mpi_abort(mpi_comm_world,ierr,ierr)

    if (rank==0) then
      write(*,*) 'slab decomposition dist_init running on',nodes,'nodes'
    endif

  end subroutine mpi_initialize

!-------------------------------------------------------------------!

  subroutine fftw(command)
!! calculate fftw transform
!! 0 ends fftw subprogram, 1 starts forward fft, -1 starts backwards
    implicit none
!! may have to root around to find this file.  It's included with the fftw source.
    include 'fftw_f77.i'

    integer(4), parameter :: order=FFTW_NORMAL_ORDER ! FFTW_TRANSPOSED_ORDER

    integer(4) :: i
    integer(4) :: command

#ifdef DEBUG_LOW
    do i=0,nodes-1
      if (rank == i) print *,'starting fftw',rank
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif

! initialize plan variables for fftw

    if (firstfftw) then
      call rfftw3d_f77_mpi_create_plan(plan,mpi_comm_world,nc, &
            nc,nc, FFTW_REAL_TO_COMPLEX, FFTW_MEASURE)
      call rfftw3d_f77_mpi_create_plan(iplan,mpi_comm_world,nc, &
            nc,nc, FFTW_COMPLEX_TO_REAL, FFTW_MEASURE)
#ifdef DEBUG_LOW
      print *,'finished initialization of fftw',rank
#endif
      firstfftw=.false.
    endif

! giver

    if (command /= 0) then

      if (command > 0) then
        call rfftwnd_f77_mpi(plan,1,slab,slab_work,1,order)
      else
        call rfftwnd_f77_mpi(iplan,1,slab,slab_work,1,order)
        slab=slab/ncr**3
      endif

    else

! if command = 0 we delete the plans

      call rfftwnd_f77_mpi_destroy_plan(iplan)
      call rfftwnd_f77_mpi_destroy_plan(plan)
    endif

  end subroutine fftw

!-------------------------------------------------------------------!

  subroutine writeparams
    implicit none

    write(*,*) 'nodes   ', nodes
    write(*,*) 'nc      ', nc
    write(*,*) 'np      ', np
    write(*,*) 'np_buffer',np_buffer
    write(*,*) 'max_np   ',max_np
    write(*,*)
    write(*,*) 'n_s      ',ns
    write(*,*) 'sigma_8  ',s8
    write(*,*) 'omega_m  ',omegam
    write(*,*) 'omega_l  ',omegal
    write(*,*)
    write(*,*) 'box      ',box
    write(*,*) 'redshift ',redshift
    write(*,*)

    return
  end subroutine writeparams

!-------------------------------------------------------------------!

  !! Dark matter data
  !! xvp(1:3) = x,y,z in grid units from 0 to nc
  !! xvp(4:6) = vx,vy,vz in km/s
  subroutine writedm
    implicit none
    character*60 :: fn

    !! for more than 1+1e6 processes this will have to be increased
    character(len=6) :: rank_s

    integer :: i,j 
    integer :: tag,blocksize,num_writes,nplow,nphigh,nptot
    integer :: mpisend_max
    integer :: rank_node,rank_io,pm_rank,np_in(cpu_node)
    integer, dimension(mpi_status_size) :: status

    blocksize=(2047*1024*1024)/24
    mpisend_max=80000000

    !! calculate local rank on node and local root process 
    rank_node=mod(rank,cpu_node)
    rank_io=rank-rank_node
    pm_rank=rank/cpu_node

    np_in(1)=np_local

    !! Tally up total number of particles on node
    do i=1,cpu_node-1
      tag=i
      if (rank_node==0) then
        call mpi_recv(np_in(i+1),1,mpi_integer,rank+i,tag, &
                      mpi_comm_world,status,ierr)
      elseif(rank_node==i) then
        call mpi_send(np_local,1,mpi_integer,rank_io,tag, &
                      mpi_comm_world,ierr)
      endif
      call mpi_barrier(mpi_comm_world,ierr)
    enddo

    !! local root process writes first
    if (rank_node==0) then
      nptot=sum(np_in)
      write(rank_s,'(i6)') pm_rank
      rank_s=adjustl(rank_s)
      fn=ic_path//'xvp'//rank_s(1:len_trim(rank_s))//'.init'
      write(*,*) 'Writing ',fn
      open(11,file=fn,form=IOform)
      write(11) nptot
      num_writes=np_in(1)/blocksize+1
      do i=1,num_writes
        nplow=(i-1)*blocksize+1
        nphigh=min(i*blocksize,np_in(1))
        write(11) xvp(:,nplow:nphigh)
      enddo
    endif
    call mpi_barrier(mpi_comm_world,ierr)

    !! now other processes transmit data to root process
    do i=1,cpu_node-1
      tag=i
      if (rank_node==0) then
        if (np_in(i+1) > mpisend_max) then
          call mpi_recv(xvp(1,1),6*mpisend_max,mpi_real,rank+i,tag, &
                        mpi_comm_world,status,ierr)
          call mpi_recv(xvp(1,mpisend_max+1),6*(np_in(i+1)-mpisend_max),mpi_real, &
                        rank+i,tag+100,mpi_comm_world,status,ierr)
        else
          call mpi_recv(xvp,6*np_in(i+1),mpi_real,rank+i,tag, &
                        mpi_comm_world,status,ierr)
        endif
        do j=1,np_in(i+1)
          xvp(3,j)=min(xvp(3,j)+i*nc_slab,real(cpu_node*nc_slab)-eps)
        enddo
        num_writes=np_in(i+1)/blocksize+1
        do j=1,num_writes
          nplow=(j-1)*blocksize+1
          nphigh=min(j*blocksize,np_in(i+1))
          write(11) xvp(:,nplow:nphigh)
        enddo
      elseif(rank_node==i) then
        if (np_local > mpisend_max) then
          call mpi_send(xvp(1,1),6*mpisend_max,mpi_real,rank_io,tag, &
                        mpi_comm_world,ierr)
          call mpi_send(xvp(1,mpisend_max+1),6*(np_local-mpisend_max),mpi_real, &
                        rank_io,tag+100,mpi_comm_world,ierr)
        else  
          call mpi_send(xvp,6*np_local,mpi_real,rank_io,tag, &
                        mpi_comm_world,ierr)
        endif
      endif
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
    if (rank_node==0) close(11)

  end subroutine  writedm

!-------------------------------------------------------------------!

  subroutine writepowerspectra
    implicit none
    integer      :: k
    real         :: kr
    character*60 :: fn

    !! Output power spectrum
    !! 1st column is k
    !! 2nd is matter p(k)
    !! 3rd is standard deviation
    !! 4th is dm p(k)
    !! 5th is standard deviation

    fn=ic_path//'pk.init'
    write(*,*) 'Writing ',fn
    open(11,file=fn,recl=500)
    do k=2,hc+1
       kr=2*pi*(k-1)/box
       write(11,*) kr,pkm(:,k),pkdm(:,k)
    enddo
    close(11)

    !! Output cmbfast power spectrum
    fn=ic_path//'pk0.init'
    write(*,*) 'Writing ',fn
    open(11,file=fn,recl=500)
    do k=2,2*nc+1
       kr=2*pi*(k-1)/(2*box)
       write(11,*) kr,power(kr,1,2),power(kr,1,3)
    enddo
    close(11)

    return
  end subroutine writepowerspectra

!!------------------------------------------------------------------!!

  subroutine transferfnc
    implicit none
    integer :: i,k
    real    :: kr,kmax
    real*8  :: v8

    if (rank == 0) then
      !! Get transfer function from CMBFAST
      write(*,*) 'Reading ',fntf
      open(11,file=fntf)
      read(11,*) tf
      close(11)

      !! Compute \Delta^2
      do k=1,nk
         kr     =tf(1,k)
         tf(2,k)=kr**(3+ns)*tf(2,k)**2/(2*pi**2)
         tf(3,k)=kr**(3+ns)*tf(3,k)**2/(2*pi**2)
      enddo

      !! Compute dk
      tf(4,1)=tf(1,2)/2
      do k=2,nk-1
         tf(4,k)=(tf(1,k+1)-tf(1,k-1))/2
      enddo
      tf(4,nk)=tf(1,nk)-tf(1,nk-1)

      !! Compute variance in 8 h^-1 Mpc spheres
      v8=0
      kmax=2*pi*sqrt(3.)*hc/box
      do k=1,nk
         if (tf(1,k) .gt. kmax) exit
         v8=v8+tf(2,k)*tophat(tf(1,k)*8)**2*tf(4,k)/tf(1,k)
      enddo

      !! Normalize to \sigma_8
      !! Include growth factor
      tf(2:3,:)=tf(2:3,:)*(s8**2/v8)*Dgrow(scalefactor)**2
    endif

    call mpi_bcast(tf,5*nk,mpi_real,0,mpi_comm_world,ierr) 
   
    !! tf(1,i) stores k
    !! tf(2,i) stores \Delta^2_m
    !! tf(3,i) stores \Delta^2_b
    !! tf(4,i) stores dk

    return
  end subroutine transferfnc

!------------------------------------------------------------------!

  subroutine noisemap
    implicit none
!! for more than 1+1e6 processes rank_s will have to be increased
    character(len=6) :: rank_s
    integer      :: i,j,k,seedsize
    real         :: x,x1,x2
    character*60 :: fn
    integer,allocatable,dimension(:) :: iseed

    write(rank_s,'(i6)') rank
    rank_s=adjustl(rank_s)

    call random_seed
    call random_seed(size=seedsize)
    allocate(iseed(seedsize))

    !! Either generate seeds or read from a file
    if (generate_seeds) then
       do k=1,seedsize
          do i=0,rank
            call random_number(x)
          enddo
          iseed(k)=int(x*huge(0))+rank
       enddo
    else
       fn=ic_path//'seed'//rank_s(1:len_trim(rank_s))//'.init'
       open(11,file=fn)
       do k=1,seedsize
          read(11,*) j,iseed(k)
       enddo
       close(11)
    endif
    call random_seed(put=iseed(1:seedsize))

    !! Generate random reals between 0 and 1
     call random_number(slab)

#ifdef DEBUG
    print *,'random sample',rank,slab(1,1,1),slab(nc+2,nc,nc_slab)
    x1=100000.0
    x2=-100000.0
    do k=1,nc_slab
      do j=1,nc
        do i=1,nc
          x1=min(x1,slab(i,j,k))
          x2=max(x2,slab(i,j,k))
        enddo
      enddo
    enddo
    print *,'rand min max',rank,x1,x2
#endif

    !! Write seeds to disk
    fn=ic_path//'seed'//rank_s(1:len_trim(rank_s))//'.init'
    open(11,file=fn)
    do k=1,seedsize
       write(11,*) k,iseed(k) 
    enddo
    close(11)

    deallocate(iseed)
    
    !! Generate random Gaussian numbers
    do k=1,nc_slab
       do j=1,nc
          do i=1,nc,2
             x1=2*pi*slab(i,j,k)
!! this log calc chokes on 0.0  :(
             if (slab(i+1,j,k)<=0.0) slab(i+1,j,k)=0.00001
             x2=sqrt(-2*log(slab(i+1,j,k)))
             slab(i,j,k)=x2*cos(x1)
             slab(i+1,j,k)=x2*sin(x1)
          enddo
       enddo
    enddo

#ifdef DEBUG
    print *,'gaussian sample',rank,slab(1,1,1),slab(nc+2,nc,nc_slab)
    x1=100000.0
    x2=-100000.0
    do k=1,nc_slab
      do j=1,nc
        do i=1,nc
          x1=min(x1,slab(i,j,k))
          x2=max(x2,slab(i,j,k))
        enddo
      enddo
    enddo
    print *,'min max',rank,x1,x2
#endif

    !! Forward FFT white noise field
    call fftw(1)

#ifdef DEBUG
    print *,'fourier gaussians sample',rank,slab(1,1,1),slab(nc+2,nc,nc_slab)
#endif

    !! Improve sampling at large scales (small k)
    !! Use only for diagnostic purposes
    if (normalize_noise) call normalizenoise

    return
  end subroutine noisemap

!-------------------------------------------------------------------!

  subroutine normalizenoise
    implicit none
    real, parameter :: err=0.1

    integer :: i,j,k,n
    integer :: k1,k2,kg
    real    :: kr,kx,ky,kz
    real    :: pk,wk,norm
    real*8, dimension(2,hc,nc)  :: A,Asum

    A=0.0
    Asum=0.0
    do k=1,nc_slab
       kg=k+nc_slab*rank
       if (kg .lt. hc+2) then
          kz=kg-1
       else
          kz=kg-1-nc
       endif
       do j=1,nc
          if (j .lt. hc+2) then
             ky=j-1
          else
             ky=j-1-nc
          endif
          do i=1,nc+2,2
             kx=(i-1)/2
             kr=sqrt(kx**2+ky**2+kz**2)
             n=int(kr)
             if (n .gt. 0 .and. n .le. hc) then
                A(1,n,kg)=A(1,n,kg)+sum(slab(i:i+1,j,k)**2)
                A(2,n,kg)=A(2,n,kg)+1
             endif
          enddo
       enddo
    enddo
 
    !! Merge to master node
    call mpi_reduce(A,Asum,2*hc*nc,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr) 

    !! Compute normalization
    if (rank == 0) then
      do k=2,nc
        Asum(:,:,1)=Asum(:,:,1)+Asum(:,:,k)
      enddo
      do k=1,hc
        norm=Asum(1,k,1)/Asum(2,k,1)/ncr**3
        Asum(1,k,1)=1
        if (norm .gt. 1+err) Asum(1,k,1)=sqrt(norm/(1+err))
        if (norm .lt. 1-err) Asum(1,k,1)=sqrt(norm/(1-err))
        write(*,*) k,real(norm/Asum(1,k,1)**2)
      enddo
    endif

    !! Broadcast back to all nodes
    call mpi_bcast(Asum,2*hc*nc,mpi_double_precision,0,mpi_comm_world,ierr)

    do k=1,nc_slab
       kg=k+nc_slab*rank
       if (kg .lt. hc+2) then
          kz=kg-1
       else
          kz=kg-1-nc
       endif
       do j=1,nc
          if (j .lt. hc+2) then
             ky=j-1
          else
             ky=j-1-nc
          endif
          do i=1,nc+2,2
             kx=(i-1)/2
             kr=sqrt(kx**2+ky**2+kz**2)
             n=int(kr)
             if (n .gt. 0 .and. n .le. hc) then
                slab(i:i+1,j,k)=slab(i:i+1,j,k)/Asum(1,n,1)
             endif
          enddo
       enddo
    enddo

    return
  end subroutine normalizenoise

!-----------------------------------------------------------------------!

  subroutine deltafield
    implicit none
    integer :: i,j,k,kg
    real    :: kr,kx,ky,kz
    real    :: powb,powm
    real    :: d,dmin,dmax,dmint,dmaxt
    real*8  :: dsum,dvar,dsumt,dvart

    !! Interpolate \Delta^2 to grid
    !! Determine Fourier modes \delta(k)

    do k=1,nc_slab
       kg=k+nc_slab*rank
       if (kg .lt. hc+2) then
         kz=kg-1
       else
         kz=kg-1-nc
       endif
       do j=1,nc
          if (j .lt. hc+2) then
             ky=j-1
          else
             ky=j-1-nc
          endif
          do i=1,nc+2,2
             kx=(i-1)/2
             kr=sqrt(kx**2+ky**2+kz**2)
             if (kr .eq. 0) then
                slab(i:i+1,j,k)=0
             else
                powm=power(2*pi*kr/box,1,2)/(4*pi*kr**3)
                powb=power(2*pi*kr/box,1,3)/(4*pi*kr**3)
                slab(i:i+1,j,k)=sqrt(powm*ncr**3)*slab(i:i+1,j,k)
             endif
          enddo
       enddo
    enddo

#ifdef DEBUG
    print *,'fourier delta sample',rank,slab(1,1,1),slab(nc+2,nc,nc_slab)
#endif

    !! Compute matter power spectrum
    call powerspectrum(pkm)

    do k=1,nc_slab
      do j=1,nc
        do i=1,nc+2
          delta(i,j,k)=slab(i,j,k)
        enddo
      enddo
    enddo

    !! Calculate Delta statistics

    call fftw(-1)

    dmin=0
    dmax=0
    dsum=0
    dvar=0
    do k=1,nc_slab
       do j=1,nc
          do i=1,nc
             d=slab(i,j,k)
             dsum=dsum+d
             dvar=dvar+d*d
             dmin=min(dmin,d)
             dmax=max(dmax,d)
          enddo
       enddo
    enddo

    call mpi_reduce(dsum,dsumt,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    call mpi_reduce(dvar,dvart,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    call mpi_reduce(dmin,dmint,1,mpi_real,mpi_min,0,mpi_comm_world,ierr)
    call mpi_reduce(dmax,dmaxt,1,mpi_real,mpi_max,0,mpi_comm_world,ierr)

    if (rank==0) then
      dsum=dsumt/nc**3
      dvar=sqrt(dvart/ncr**3)
      write(*,*)
      write(*,*) 'Delta min    ',dmint
      write(*,*) 'Delta max    ',dmaxt
      write(*,*) 'Delta sum ',real(dsum)
      write(*,*) 'Delta var ',real(dvar)
      write(*,*)
    endif

    return
  end subroutine deltafield

!---------------------------------------------------------------------!

  subroutine potentialfield
    implicit none

    integer :: i,j,k
    integer :: im,ip,ig,jm,jp,jg,km,kp,kg
    real    :: r,x,y,z
    real    :: kr,ksq,kx,ky,kz
    real    :: phi8,phi8tot

    !! Construct uncorrected potential kernel in Fourier space
    do k=1,nc_slab
       kg=k+nc_slab*rank
       if (kg .lt. hc+2) then
          kz=kg-1
       else
          kz=kg-1-nc
       endif
       kz=2*sin(pi*kz/ncr)
       do j=1,nc
          if (j .lt. hc+2) then
             ky=j-1
          else
             ky=j-1-nc
          endif
          ky=2*sin(pi*ky/ncr)
          do i=1,nc+2,2
             ip=i+1
             kx=(i-1)/2
             kx=2*sin(pi*kx/ncr)
             ksq=kx**2+ky**2+kz**2
             if (ksq .eq. 0) then
                slab(i:ip,j,k)=0
             else
                slab(i,j,k)=-4*pi/ksq
                slab(ip,j,k)=0
             endif
          enddo
       enddo
    enddo

#ifdef DEBUG_KC
    phi8=0.0
    phi8tot=0.0
    do k=1,nc_slab
      do j=1,nc
        do i=1,nc+2
          phi8=phi8+delta(i,j,k)
        enddo
      enddo
    enddo
    call mpi_reduce(phi8,phi8tot,1,mpi_real,mpi_sum,0,mpi_comm_world,ierr)
    if (rank==0) print *,'total delta=',phi8tot
#endif


    if (correct_kernel) then

       !! Inverse FFT potential kernel
       call fftw(-1) 

       !! Compute offset at r=8
       phi8=0.0
       if (nc_slab < 9) then
         print *,'warning: mesh too small in potential kernel correction'
         call mpi_abort(mpi_comm_world,ierr,ierr)
       endif

       if (rank == 0) phi8=slab(9,1,1)+slab(1,9,1)+slab(1,1,9)+slab(nc-7,1,1)+ &
                           slab(1,nc-7,1)

       if (rank == nodes-1) phi8=phi8+slab(1,1,nc_slab-7)

       call mpi_reduce(phi8,phi8tot,1,mpi_real,mpi_sum,0,mpi_comm_world,ierr)

       if (rank == 0) phi8=phi8tot/6.0      

       call mpi_bcast(phi8,1,mpi_real,0,mpi_comm_world,ierr)

       !! Construct Ewald potential kernel in real space
       do k=1,nc_slab
          kg=k+nc_slab*rank
          if (kg .lt. hc+2) then
             z=kg-1
          else
             z=kg-1-nc
          endif
          do j=1,nc
             if (j .lt. hc+2) then
                y=j-1
             else
                y=j-1-nc
             endif
             do i=1,nc
                if (i .lt. hc+2) then
                   x=i-1
                else
                   x=i-1-nc
                endif
                r=sqrt(x**2+y**2+z**2)
                if (r .gt. 8) then
                   slab(i,j,k)=slab(i,j,k)-(phi8+1.0/8.0)
                else
                   if (r .eq. 0) then
                      slab(i,j,k)=-2.5
                   else
                      slab(i,j,k)=-1/r
                   endif
                endif
             enddo
          enddo
       enddo

#ifdef DEBUG_KC
!       if (rank==0) then
!         do i=1,16
!           print *,rank,slab(i,i,i)
!         enddo
!       endif
       print *,'rank',rank,'sum kern',sum(slab)
#endif 

       !! Forward FFT potential kernel
       call fftw(1)

    endif

#ifdef DEBUG_KC
    phi8=0.0
    phi8tot=0.0
    do k=1,nc_slab
      do j=1,nc
        do i=1,nc+2  
          phi8=phi8+slab(i,j,k)
        enddo
      enddo
    enddo
    call mpi_reduce(phi8,phi8tot,1,mpi_real,mpi_sum,0,mpi_comm_world,ierr)
    if (rank==0) print *,'total slab=',phi8tot
    phi8=0.0
    phi8tot=0.0
    do k=1,nc_slab
      do j=1,nc
        do i=1,nc+2
          phi8=phi8+delta(i,j,k)
        enddo
      enddo
    enddo
    call mpi_reduce(phi8,phi8tot,1,mpi_real,mpi_sum,0,mpi_comm_world,ierr)
    if (rank==0) print *,'total delta=',phi8tot
#endif

    !! Complex multiply density field with potential kernel
    do k=1,nc_slab
       do j=1,nc
          do i=1,nc+2,2
             slab(i:i+1,j,k)=delta(i:i+1,j,k)*slab(i,j,k)
          enddo
       enddo
    enddo

    !! Inverse FFT potential field
    call fftw(-1)

    phi(:,:,1:nc_slab)=slab(:nc,:,:) 

    call phi_buffer

    return
  end subroutine potentialfield

!!------------------------------------------------------------------!!

  subroutine phi_buffer 

    implicit none
  
    integer(kind=4) :: status(MPI_STATUS_SIZE)
    integer(kind=4) :: tag,ierr,rank_to,rank_from,buf_size

    buf_size=2*nc*nc

    tag=42

    phi(:,:,nc_slab+1:nc_slab+2)=phi(:,:,1:2)

    rank_to=mod(rank+nodes-1,nodes)
    rank_from=mod(rank+1,nodes)

    call mpi_sendrecv_replace(phi(:,:,nc_slab+1:nc_slab+2),buf_size,MPI_REAL, &
        rank_to,tag,rank_from,tag,MPI_COMM_WORLD,status,ierr)

    tag=54

    phi(:,:,-1:0)=phi(:,:,nc_slab-1:nc_slab)
  
    rank_to=mod(rank+1,nodes)
    rank_from=mod(rank+nodes-1,nodes)

    call mpi_sendrecv_replace(phi(:,:,-1:0),buf_size,MPI_REAL, &
        rank_to,tag,rank_from,tag,MPI_COMM_WORLD,status,ierr)

  end subroutine phi_buffer 

!!------------------------------------------------------------------!!

  subroutine den_buffer 

    implicit none
  
    integer(kind=4) :: status(MPI_STATUS_SIZE)
    integer(kind=4) :: tag,ierr,rank_to,rank_from,buf_size
    real*8 :: sum_dm,sum_dm_local

    buf_size=nc*nc

    tag=42

    rank_to=mod(rank+nodes-1,nodes)
    rank_from=mod(rank+1,nodes)

#ifdef DEBUG_DB
    call mpi_barrier(mpi_comm_world,ierr)
    print *,'a',rank,rank_to,rank_from,sum(phi_buf)
    sum_dm_local=sum(phi_buf)
    call mpi_reduce(sum_dm_local,sum_dm,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    if (rank == 0) print *,'downward out=',sum_dm
#endif
    call mpi_sendrecv_replace(phi(:,:,0),buf_size,MPI_REAL, &
        rank_to,tag,rank_from,tag,MPI_COMM_WORLD,status,ierr)

    phi(:,:,nc_slab)=phi(:,:,nc_slab)+phi(:,:,0)

#ifdef DEBUG_DB
    call mpi_barrier(mpi_comm_world,ierr)
    print *,'b',rank,rank_to,rank_from,sum(phi_buf)
    sum_dm_local=sum(phi_buf)
    call mpi_reduce(sum_dm_local,sum_dm,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    if (rank == 0) print *,'upward in=',sum_dm
#endif

    tag=54

#ifdef DEBUG_DB
    call mpi_barrier(mpi_comm_world,ierr)
    print *,'c',rank,rank_to,rank_from,sum(phi_buf)
    sum_dm_local=sum(phi_buf)
    call mpi_reduce(sum_dm_local,sum_dm,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    if (rank == 0) print *,'upward out=',sum_dm
#endif

    rank_to=mod(rank+1,nodes)
    rank_from=mod(rank+nodes-1,nodes)

    call mpi_sendrecv_replace(phi(:,:,nc_slab+1),buf_size,MPI_REAL, &
        rank_to,tag,rank_from,tag,MPI_COMM_WORLD,status,ierr)

#ifdef DEBUG_DB
    call mpi_barrier(mpi_comm_world,ierr)
    print *,'d',rank,rank_to,rank_from,sum(phi_buf)
    sum_dm_local=sum(phi_buf)
    call mpi_reduce(sum_dm_local,sum_dm,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    if (rank == 0) print *,'downward in=',sum_dm
#endif

    phi(:,:,1)=phi(:,:,1)+phi(:,:,nc_slab+1)

  end subroutine den_buffer 

!!------------------------------------------------------------------!!

  subroutine dm
    implicit none
    integer :: i,j,k
    integer :: i1,j1,k1,ip,im,jp,jm,kp,km
    real    :: d,dmin,dmax,dmint,dmaxt,df(3)
    real*8  :: dsum,dvar,dsumt,dvart,x(6),sum_dm,sum_dm_local
    real    :: vf,v(3)

    vf=vfactor(scalefactor)

    !! Finite-difference phi to get displacement field
    !! Displace particles
    do k=1,np_slab
       k1=(nc/np)*(k-1)+1
       kp=k1+1
       km=k1-1
       do j=1,np
          j1=(nc/np)*(j-1)+1
          jp=mod(j1,nc)+1
          jm=mod(j1-2+nc,nc)+1
          do i=1,np
             i1=(nc/np)*(i-1)+1
             ip=mod(i1,nc)+1
             im=mod(i1-2+nc,nc)+1
             df(1)=(phi(im,j1,k1)-phi(ip,j1,k1))/2/(4*pi)
             df(2)=(phi(i1,jm,k1)-phi(i1,jp,k1))/2/(4*pi)
             df(3)=(phi(i1,j1,km)-phi(i1,j1,kp))/2/(4*pi)
#ifdef GRID_TEST
             xvp_grid(1,i,j,k)=i1 -1 !.5
             xvp_grid(2,i,j,k)=j1 -1 !.5
             xvp_grid(3,i,j,k)=k1 -1 !.5
             xvp_grid(4:,i,j,k)=0.0
#else
             xvp_grid(1,i,j,k)=mod(df(1)+(i1-0.5)+ncr,ncr)
             xvp_grid(2,i,j,k)=mod(df(2)+(j1-0.5)+ncr,ncr)
             xvp_grid(3,i,j,k)=df(3)+(k1-0.5)
             xvp_grid(4:,i,j,k)=vf*df
#endif
          enddo
       enddo
    enddo

    np_local=np_slab*np*np

#ifdef DEBUG
    print *,'first particle',xvp(:,1)
    x=0.0
    do i=1,np_local
      x=x+xvp(:,i)
    enddo
    x=x/real(np_local)
    do i=0,nodes-1
      if (rank==i) print *,rank,'avg position/velocity =',x
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif

    !! Pass particles that have been displaced out of local mesh
    call pass_particles

!! assign velocities here!!

    !! Initialized density field to be zero
    !! Re-using phi array
    do k=-1,nc_slab+2
       phi(:,:,k)=0.0
    enddo

#ifdef DEBUG
    sum_dm_local=sum(phi)
    print *,'rank',rank,'sum phi clean=',sum_dm_local
    call mpi_reduce(sum_dm_local,sum_dm,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    if (rank == 0) print *,'phi total clean mass=',sum_dm
    sum_dm_local=0.0
    do k=1,nc_slab
      do j=1,nc
        do i=1,nc
          sum_dm_local=sum_dm_local+phi(i,j,k)
        enddo
      enddo
    enddo
    print *,'rank',rank,'sum manual phi clean=',sum_dm_local
#endif

#ifdef DEBUG
    do k=-1,nc_slab+2
      do j=1,nc
        do i=1,nc
          if (phi(i,j,k) < 0.0) print *,rank,'phi<0.0',i,j,k,phi(i,j,k)
        enddo
      enddo
    enddo
    if (rank==0) print *,'finished initial phi check'
#endif

    !! Assign masses to grid to compute dm power spectrum
    call cicmass

#ifdef DEBUG
    do k=-1,nc_slab+2
      do j=1,nc
        do i=1,nc
          if (phi(i,j,k) < 0.0) print *,rank,'phi<0.0',i,j,k,phi(i,j,k)
        enddo
      enddo
    enddo

    sum_dm_local=sum(phi)
    print *,'rank',rank,'sum den b4 buffer local=',sum_dm_local
    call mpi_reduce(sum_dm_local,sum_dm,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    if (rank == 0) print *,'phi total mass=',sum_dm
    sum_dm_local=0.0
    do k=1,nc_slab
      do j=1,nc
        do i=1,nc
          sum_dm_local=sum_dm_local+phi(i,j,k)
        enddo
      enddo
    enddo
    print *,'rank',rank,'sum den b4 buffer local manual=',sum_dm_local
#endif

    !! have to accumulate buffer density 
    call den_buffer 

#ifdef DEBUG
    sum_dm_local=sum(phi(:,:,1:nc_slab))
    print *,'rank',rank,'sum den b4 copy local=',sum_dm_local
    call mpi_reduce(sum_dm_local,sum_dm,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    if (rank == 0) print *,'b4 copy total mass=',sum_dm
    sum_dm_local=0.0
    do k=1,nc_slab
      do j=1,nc
        do i=1,nc
          sum_dm_local=sum_dm_local+phi(i,j,k)
        enddo
      enddo
    enddo
    print *,'rank',rank,'sum den b4 copy local manual=',sum_dm_local
#endif

    slab(:nc,:,:)=phi(:,:,1:nc_slab)
    slab(nc+1:,:,:)=0.0

!    sum_dm_local=sum(slab) 
!    print *,'rank',rank,'sum dm local=',sum_dm_local
!    call mpi_reduce(sum_dm_local,sum_dm,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
!    if (rank == 0) print *,'DM total mass=',sum_dm
    sum_dm_local=0.0
    do k=1,nc_slab
      do j=1,nc
        do i=1,nc
          if (slab(i,j,k) < 0.0) print *,'rank dm < 0.0',i,j,k,slab(i,j,k)
          sum_dm_local=sum_dm_local+slab(i,j,k)
        enddo
      enddo
    enddo
    call mpi_reduce(sum_dm_local,sum_dm,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    if (rank == 0) print *,'DM total mass =',sum_dm

    !! Convert dm density field to delta field
    dmin=0
    dmax=0
    dsum=0
    dvar=0

    do k=1,nc_slab
       do j=1,nc
          do i=1,nc
             slab(i,j,k)=slab(i,j,k)-1.0
             d=slab(i,j,k)
             dsum=dsum+d
             dvar=dvar+d*d
             dmin=min(dmin,d)
             dmax=max(dmax,d)
          enddo
       enddo
    enddo

    call mpi_reduce(dsum,dsumt,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    call mpi_reduce(dvar,dvart,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    call mpi_reduce(dmin,dmint,1,mpi_real,mpi_min,0,mpi_comm_world,ierr)
    call mpi_reduce(dmax,dmaxt,1,mpi_real,mpi_max,0,mpi_comm_world,ierr)

    if (rank==0) then
      dsum=dsumt/nc**3
      dvar=sqrt(dvart/ncr**3)
      write(*,*)
      write(*,*) 'DM min    ',dmint
      write(*,*) 'DM max    ',dmaxt
      write(*,*) 'DM delta sum ',real(dsum)
      write(*,*) 'DM delta var ',real(dvar)
      write(*,*)
    endif
 
    !! Forward FFT dm delta field
    call fftw(1)

    !! Compute dm power spectrum
    call powerspectrum(pkdm)

    return
  end subroutine dm

!------------------------------------------------------------!

  subroutine pass_particles
!! The smart way to ascertain the degree of particle transfer
!! would be to investigate the displacement field and then 
!! dynamically allocate the expected memory for transitory 
!! particles -- currently hardwired in a constant! (np_buffer)
    implicit none

    integer i,pp,np_buf,np_exit,npo,npi
    real lb,ub
    integer*8 :: nplr,np_final
    integer, dimension(mpi_status_size) :: status,sstatus,rstatus
    integer :: rank_to,rank_from,tag,srequest,rrequest,sierr,rierr

    lb=0.0
    ub=real(nc_slab)

#ifdef DEBUG
    do i=0,nodes-1
      if (rank==i) print *,rank,'np_local=',np_local
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif 

    np_buf=0
    pp=1
    do
      if (pp > np_local) exit
      if (xvp(3,pp) < lb .or. xvp(3,pp) >= ub ) then
        np_buf=np_buf+1
        if (np_buf > np_buffer) then
          print *,rank,'np_buffer =',np_buffer,'exceeded - np_buf =',np_buf
          call mpi_abort(mpi_comm_world,ierr,ierr)
        endif 
        xp_buf(:,np_buf)=xvp(:,pp)
        xvp(:,pp)=xvp(:,np_local)
        np_local=np_local-1
        cycle 
      endif
      pp=pp+1
    enddo
 
    call mpi_reduce(np_buf,np_exit,1,mpi_integer,mpi_sum,0, &
                    mpi_comm_world,ierr) 

#ifdef DEBUG
    do i=0,nodes-1
      if (rank==i) print *,rank,'np_exit=',np_buf
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif 

    if (rank == 0) print *,'total exiting particles =',np_exit

! pass +z

    tag=11 
    npo=0
    pp=1
    do 
      if (pp > np_buf) exit
      if (xp_buf(3,pp) >= ub) then
        npo=npo+1
        send_buf((npo-1)*6+1:npo*6)=xp_buf(:,pp)
        xp_buf(:,pp)=xp_buf(:,np_buf)
        np_buf=np_buf-1
        cycle
      endif
      pp=pp+1
    enddo

#ifdef DEBUG
    do i=0,nodes-1
      if (rank==i) print *,rank,'np_out=',npo
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif 
    
    rank_to=mod(rank+1,nodes)
    rank_from=mod(rank+nodes-1,nodes)

    npi = npo

    call mpi_sendrecv_replace(npi,1,mpi_integer,rank_to,&
                              tag,rank_from,tag,mpi_comm_world, &
                              status,ierr) 

    call mpi_isend(send_buf,npo*6,mpi_real,rank_to, &
                   tag,mpi_comm_world,srequest,sierr)
    call mpi_irecv(recv_buf,npi*6,mpi_real,rank_from, &
                   tag,mpi_comm_world,rrequest,rierr)
    call mpi_wait(srequest,sstatus,sierr)
    call mpi_wait(rrequest,rstatus,rierr)

    do pp=1,npi
      xp_buf(:,np_buf+pp)=recv_buf((pp-1)*6+1:pp*6)
      xp_buf(3,np_buf+pp)=max(xp_buf(3,np_buf+pp)-ub,lb)
    enddo

#ifdef DEBUG
    do i=0,nodes-1
      if (rank==i) print *,rank,'z+ np_local=',np_local
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif 

    pp=1
    do 
      if (pp > npi) exit 
      if (xp_buf(3,np_buf+pp) >= lb .and. xp_buf(3,np_buf+pp) < ub ) then
        np_local=np_local+1
        xvp(:,np_local)=xp_buf(:,np_buf+pp)
        xp_buf(:,np_buf+pp)=xp_buf(:,np_buf+npi)
        npi=npi-1
        cycle
      endif
      pp=pp+1
    enddo
   
    np_buf=np_buf+npi

#ifdef DEBUG
    do i=0,nodes-1
      if (rank==i) print *,rank,'z+ np_exit=',np_buf,np_local
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif 

! pass -z

    tag=12
    npo=0
    pp=1
    do
      if (pp > np_buf) exit
      if (xp_buf(3,pp) < lb) then
        npo=npo+1
        send_buf((npo-1)*6+1:npo*6)=xp_buf(:,pp)
        xp_buf(:,pp)=xp_buf(:,np_buf)
        np_buf=np_buf-1
        cycle 
      endif
      pp=pp+1
    enddo

#ifdef DEBUG
    do i=1,np_buf
      print *,xp_buf(:,i)
    enddo
    do i=0,nodes-1
      if (rank==i) print *,rank,'particles in buffer before down pass=',np_buf
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif 

    npi = npo

    rank_to=mod(rank+nodes-1,nodes)
    rank_from=mod(rank+1,nodes)

    call mpi_sendrecv_replace(npi,1,mpi_integer,rank_to, &
                              tag,rank_from,tag,mpi_comm_world, &
                              status,ierr)

    call mpi_isend(send_buf,npo*6,mpi_real,rank_to, &
                   tag,mpi_comm_world,srequest,sierr)
    call mpi_irecv(recv_buf,npi*6,mpi_real,rank_from, &
                   tag,mpi_comm_world,rrequest,rierr)
    call mpi_wait(srequest,sstatus,sierr)
    call mpi_wait(rrequest,rstatus,rierr)

    do pp=1,npi
      xp_buf(:,np_buf+pp)=recv_buf((pp-1)*6+1:pp*6)
      xp_buf(3,np_buf+pp)=min(xp_buf(3,np_buf+pp)+ub,ub-eps)
    enddo

    pp=1
    do
      if (pp > npi) exit
      if ( xp_buf(3,np_buf+pp) >= lb .and. xp_buf(3,np_buf+pp) < ub ) then
        np_local=np_local+1
        xvp(:,np_local)=xp_buf(:,np_buf+pp)
        xp_buf(:,np_buf+pp)=xp_buf(:,np_buf+npi)
        npi=npi-1
        cycle 
      endif
      pp=pp+1
    enddo
  
    np_buf=np_buf+npi


#ifdef DEBUG
    do i=1,np_buf
      print *,xp_buf(:,i)
    enddo
    do i=0,nodes-1
      if (rank==i) print *,rank,'particles left in buffer=',np_buf
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif 

    call mpi_reduce(np_buf,np_exit,1,mpi_integer,mpi_sum,0, &
                    mpi_comm_world,ierr)

#ifdef DEBUG
    if (rank == 0) print *,'total buffered particles =',np_exit
#endif
    nplr=int(np_local,kind=8)
!    print *,rank,nplr
    np_final=0
!! For some reason this does not work properly on the lobsters...
    call mpi_reduce(nplr,np_final,1,MPI_DOUBLE_PRECISION,MPI_SUM,0, &
                    MPI_COMM_WORLD,ierr)
!    print *,'ierr',rank,ierr
    if (rank == 0) then
      print *,'total particles =',np_final
      if (np_final /= (int(np,kind=8))**3) then
        print *,'ERROR: total number of particles incorrect after passing'
      endif
    endif
 
!!  Check for particles out of bounds

    do i=1,np_local
      if (xvp(1,i) < 0 .or. xvp(1,i) >= nc .or. &
          xvp(2,i) < 0 .or. xvp(2,i) >= nc .or. &
          xvp(3,i) < 0 .or. xvp(3,i) >= nc_slab ) then
        print *,'particle out of bounds',rank,i,xvp(:,i)
      endif
    enddo

  end subroutine pass_particles

!------------------------------------------------------------!

  subroutine cicmass
    implicit none

    integer :: i
    integer :: i1,i2,j1,j2,k1,k2
    real    :: x,y,z,dx1,dx2,dy1,dy2,dz1,dz2

#ifdef MASS_CHECK 
    real :: mpc
    real(8) :: tmpc,tphi
    tmpc=0.0
#endif


    do i=1,np_local
       x=mod(xvp(1,i)-0.5+ncr,ncr)
       y=mod(xvp(2,i)-0.5+ncr,ncr)
       z=xvp(3,i)-0.5

       i1=int(x)+1
       i2=mod(i1,nc)+1
       dx1=i1-x
       dx2=1-dx1
       j1=int(y)+1
       j2=mod(j1,nc)+1
       dy1=j1-y
       dy2=1-dy1
       k1=floor(z)+1
       k2=k1+1
       dz1=k1-z
       dz2=1-dz1

#ifdef DEBUG
       if (i1 < 1 .or. i2 > nc .or. j1 < 1 .or. &
           j2 > nc .or. k1 < 0 .or. k2 > nc_slab+1) then 
         print *,'particle out of bounds',i1,i2,j1,j2,k1,k2
       endif 

       if (dx1<0 .or. dx2<0 .or. dy1<0 .or. dy2<0 ) then
         print *,'XY negative mass',rank,dx1,dx2,dy1,dy2
       endif
       if (dz1<0 .or. dz2<0) then
         print *,'z -m',z,k1,k2,dz1,dz2
       endif
#endif

       dz1=mp*dz1
       dz2=mp*dz2
#ifdef MASS_CHECK
       if (abs(i2-i1)<=1.and.abs(j2-j1)<=1.and.abs(k2-k1)<=1) then
         tphi=sum(phi(i1:i2,j1:j2,k1:k2)) 
       else
         tphi=phi(i1,j1,k1)+ &
              phi(i2,j1,k1)+&
              phi(i1,j2,k1)+&
              phi(i2,j2,k1)+&
              phi(i1,j1,k2)+&
              phi(i2,j1,k2)+&
              phi(i1,j2,k2)+&
              phi(i2,j2,k2)
       endif
#endif
       phi(i1,j1,k1)=phi(i1,j1,k1)+dx1*dy1*dz1
       phi(i2,j1,k1)=phi(i2,j1,k1)+dx2*dy1*dz1
       phi(i1,j2,k1)=phi(i1,j2,k1)+dx1*dy2*dz1
       phi(i2,j2,k1)=phi(i2,j2,k1)+dx2*dy2*dz1
       phi(i1,j1,k2)=phi(i1,j1,k2)+dx1*dy1*dz2
       phi(i2,j1,k2)=phi(i2,j1,k2)+dx2*dy1*dz2
       phi(i1,j2,k2)=phi(i1,j2,k2)+dx1*dy2*dz2
       phi(i2,j2,k2)=phi(i2,j2,k2)+dx2*dy2*dz2
#ifdef MASS_CHECK
       if (abs(i2-i1)<=1.and.abs(j2-j1)<=1.and.abs(k2-k1)<=1) then
         tphi=sum(phi(i1:i2,j1:j2,k1:k2))-tphi
       else
         tphi=phi(i1,j1,k1)+ &
              phi(i2,j1,k1)+&
              phi(i1,j2,k1)+&
              phi(i2,j2,k1)+&
              phi(i1,j1,k2)+&
              phi(i2,j1,k2)+&
              phi(i1,j2,k2)+&
              phi(i2,j2,k2)-&
              tphi
       endif 
       if (tphi < mp-0.001 .or. tphi > mp+0.001) then
         print *,'tphi',mpc,i,xvp(:,i),i1,i2,j1,j2,k1,k2,tphi
       endif
       mpc=dx1*dy1*dz1+dx2*dy1*dz1+dx1*dy2*dz1+ &
           dx2*dy2*dz1+dx1*dy1*dz2+dx2*dy1*dz2+ &
           dx1*dy2*dz2+dx2*dy2*dz2
       if (mpc < mp-0.001 .or. mpc > mp+0.001) then
         print *,mpc,i,xvp(:,i),i1,i2,j1,j2,k1,k2
       endif
       tmpc=tmpc+mpc
#endif
    enddo
#ifdef MASS_CHECK
    print *,'total mass allocated=',tmpc
#endif

    return
  end subroutine cicmass

!!--------------------------------------------------------------!!

  subroutine powerspectrum(pk)
    implicit none
    real, dimension(2,nc)       :: pk

    integer :: i,j,k,kg
    integer :: k1,k2
    real    :: kr,kx,ky,kz,w1,w2,pow
    real, dimension(3,nc) :: pkt
    real, dimension(3,nc) :: pktsum

    pkt=0.0
    pktsum=0.0
    !! Compute power spectrum
    do k=1,nc_slab
       kg=k+nc_slab*rank
       if (kg .lt. hc+2) then
          kz=kg-1
       else
          kz=kg-1-nc
       endif
       do j=1,nc
          if (j .lt. hc+2) then
             ky=j-1
          else
             ky=j-1-nc
          endif
          do i=1,nc+2,2
             kx=(i-1)/2
             kr=sqrt(kx**2+ky**2+kz**2)
             if (kr .ne. 0) then
                k1=ceiling(kr)
                k2=k1+1
                w1=k1-kr
                w2=1-w1
                pow=sum((slab(i:i+1,j,k)/ncr**3)**2)
                pkt(1,k1)=pkt(1,k1)+w1*pow
                pkt(2,k1)=pkt(2,k1)+w1*pow**2
                pkt(3,k1)=pkt(3,k1)+w1
                pkt(1,k2)=pkt(1,k2)+w2*pow
                pkt(2,k2)=pkt(2,k2)+w2*pow**2
                pkt(3,k2)=pkt(3,k2)+w2
             endif
          enddo
       enddo
    enddo

    !! Reduce to rank 0
    call mpi_reduce(pkt,pktsum,3*nc,mpi_real,mpi_sum,0,mpi_comm_world,ierr)

    !! Divide by weights
    !! pk(1,k) stores pk(k)
    !! pk(2,k) stores standard deviation
    if (rank == 0) then
      do k=1,nc
        if (pktsum(3,k) .eq. 0) then
          pk(:,k)=0
        else
          pk(:,k)=pktsum(1:2,k)/pktsum(3,k)
          pk(2,k)=sqrt(abs((pk(2,k)-pk(1,k)**2)/(pktsum(3,k)-1)))
          pk(1:2,k)=4*pi*(k-1)**3*pk(1:2,k)
       endif
      enddo
    endif

    call mpi_bcast(pk,2*nc,mpi_real,0,mpi_comm_world,ierr)

    return
  end subroutine powerspectrum

!!------------------------------------------------------------------!!

  function power(kr,ix,iy)
   implicit none
    real    :: kr
    integer :: ix,iy

    integer :: i,i1,i2
    real    :: x,y,x1,x2,y1,y2
    real    :: power

    i1=1
    i2=nk
    do while (i2-i1 .gt. 1)
       i=(i1+i2)/2
       if (kr .gt. tf(ix,i)) then
          i1=i
       else
          i2=i
       endif
    enddo

    x1=log(tf(ix,i1))
    y1=log(tf(iy,i1))
    x2=log(tf(ix,i2))
    y2=log(tf(iy,i2))
    x =log(kr)
    y =y1+(y2-y1)*(x-x1)/(x2-x1)
    power=exp(y)
    
    return
  end function power

  function Dgrow(a)
    implicit none
    real, parameter :: om=omegam
    real, parameter :: ol=omegal
    real a
    real Dgrow

    real g,ga,hsq,oma,ola

    hsq=om/a**3+(1-om-ol)/a**2+ol
    oma=om/(a**3*hsq)
    ola=ol/hsq
    g=2.5*om/(om**(4./7)-ol+(1+om/2)*(1+ol/70))
    ga=2.5*oma/(oma**(4./7)-ola+(1+oma/2)*(1+ola/70))

    Dgrow=a*ga/g
    return
  end function Dgrow

  function vfactor(a)
    implicit none
    real :: a

    real :: H,km,lm
    real :: vfactor

    lm=omegal/omegam
    km=(1-omegam-omegal)/omegam
    H=2/(3*sqrt(a**3))*sqrt(1+a*km+a**3*lm)
    vfactor=a**2*H

    return
  end function vfactor

  function tophat(x)
    implicit none
    real :: x,tophat

    if (x .ne. 0) then
       tophat=3*(sin(x)-cos(x)*x)/x**3
    else
       tophat=1
    endif

    return
  end function tophat

end program dist_init 
